from unittest import TestCase
from artof_utils.helpers.hardware import parse_gps_fix_number


class TestHelpersHardware(TestCase):
    def test_parse_gps_fix_number(self):
        self.assertEqual(parse_gps_fix_number(0), 'Not valid')
        self.assertEqual(parse_gps_fix_number(1), 'GPS Fix')
        self.assertEqual(parse_gps_fix_number(2), 'Differential GPS')
        self.assertEqual(parse_gps_fix_number(3), 'Not applicable')
        self.assertEqual(parse_gps_fix_number(4), 'RTK Fix')
        self.assertEqual(parse_gps_fix_number(5), 'RTK Float')
        self.assertEqual(parse_gps_fix_number(6), 'Not valid')
        self.assertEqual(parse_gps_fix_number(-1), 'Not valid')
        self.assertEqual(parse_gps_fix_number(None), 'Not valid')
        self.assertEqual(parse_gps_fix_number(''), 'Not valid')


from artof_utils.helpers import traject
from tests.visual_support import plot_on_top
import pandas as pd


file_name = "files/traject/complex.csv"
side = 'end'  # 'begin', 'end', 'both'
row_number = -1  # -1, 0, 1, [0, 1], [0, 2]
length = -5  # 10, 2, -5

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    df = pd.read_csv(file_name)
    df = df.apply(pd.to_numeric, errors='coerce')

    # Calculate orientation of points
    path = [[xy[0], xy[1]] for xy in zip(df['X'], df['Y'])]
    path_extended = traject.perform(traject.Operation.EXTEND, path, side=side, length=length, row_number=row_number)

    plot_on_top({"original": path, "Extended %s" % side: path_extended})



from artof_utils.helpers import array
from artof_utils.helpers import shape as shp
from shapely.geometry import Point
from copy import deepcopy
from typing import Union
import numpy as np
from enum import Enum


# Helper Classes
class TrajectPoint:
    coordinate: Point
    orientation: float

    def __init__(self, coordinate: Union[Point, list, np.ndarray], orientation: float):
        self.coordinate = coordinate
        self.orientation = orientation

    @property
    def coordinate_as_list(self):
        if isinstance(self.coordinate, Point):
            return [self.coordinate.x, self.coordinate.y]
        else:
            return self.coordinate

    def move_parallel(self, distance: float):
        dx = distance * np.cos(self.orientation)
        dy = distance * np.sin(self.orientation)
        if isinstance(self.coordinate, Point):
            self.coordinate = Point(self.coordinate.x + dx, self.coordinate.y + dy)
        elif isinstance(self.coordinate, np.ndarray) or isinstance(self.coordinate, list):
            self.coordinate = [self.coordinate[0] + dx, self.coordinate[1] + dy]

    def move_perpendicular(self, distance: float):
        perpendicular_orientation = self.orientation + np.pi / 2
        dx = distance * np.cos(perpendicular_orientation)
        dy = distance * np.sin(perpendicular_orientation)
        if isinstance(self.coordinate, Point):
            self.coordinate = Point(self.coordinate.x + dx, self.coordinate.y + dy)
        elif isinstance(self.coordinate, np.ndarray) or isinstance(self.coordinate, list):
            self.coordinate = [self.coordinate[0] + dx, self.coordinate[1] + dy]


class Operation(Enum):
    REVERSE = 'reverse'
    FLIP = 'flip'
    EXTEND = 'extend'
    SHIFT = 'shift'
    SHIFT_ALTERNATE = 'shift_alternate'
    ADD = 'add'
    REMOVE = 'remove'

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


# Helper Functions

def __parse_data(data):
    depth = array.get_depth(data)
    if depth == 3:
        return data[0]
    elif depth == 2:
        return data
    else:
        raise ValueError("Data must be 2D or 3D")


def __extend_row(row, side: str, length: float):
    if side == 'begin':
        first_point: TrajectPoint = row[0]
        first_point.move_parallel(-length)
    elif side == 'end':
        last_point: TrajectPoint = row[-1]
        last_point.move_parallel(length)
    elif side == 'both':
        __extend_row(row, 'begin', length)
        __extend_row(row, 'end', length)


def __shift_row(row, distance: float):
    for traject_point in row:
        traject_point.move_perpendicular(distance)


def __extract_rows(points: list[Point], threshold_degrees=45):
    threshold_radians = threshold_degrees * np.pi / 180
    rows: list[list[TrajectPoint]] = [[]]

    for i, point in enumerate(points):
        last_row = rows[-1]

        if i < len(points) - 1:
            orientation = shp.get_orientation(point, points[i + 1], degrees=False)
            d_orientation = orientation - (last_row[-1].orientation if len(last_row) > 0 else orientation)

            if abs(d_orientation) > threshold_radians:
                p_orientation = last_row[-1].orientation
                last_row.append(TrajectPoint(point, p_orientation))
                # Append new row
                rows.append([])
            else:
                last_row.append(TrajectPoint(point, orientation))
        else:
            # Last point
            if len(last_row) > 0:
                p_orientation = last_row[-1].orientation
                rows[-1].append(TrajectPoint(point, p_orientation))

    return rows


def get_rows(data):
    rows = __extract_rows(__parse_data(data))
    rows = [[traject_point.coordinate_as_list for traject_point in row] for row in rows]
    return rows


# Operations
def perform(operation: Operation, data, **kwargs):
    if operation == Operation.REVERSE:
        return reverse(data)
    elif operation == Operation.FLIP:
        return flip(data)
    elif operation == Operation.EXTEND:
        return extend(data, **kwargs)
    elif operation == Operation.SHIFT:
        return shift(data, **kwargs)
    elif operation == Operation.SHIFT_ALTERNATE:
        return shift_alternate(data, **kwargs)
    elif operation == Operation.ADD:
        return add(data, **kwargs)
    elif operation == Operation.REMOVE:
        return remove(data, **kwargs)
    return data


def merge_rows(rows):
    merged_rows = []
    for row in rows:
        merged_rows += [traject_point.coordinate_as_list for traject_point in row]
    return merged_rows


def reverse(data):
    depth = array.get_depth(data)
    traject_data = deepcopy(data)

    if depth == 2:
        if isinstance(traject_data, np.ndarray):
            traject_data = np.flip(traject_data)
        elif isinstance(traject_data, list):
            traject_data.reverse()
    if depth == 3:
        if isinstance(traject_data, np.ndarray):
            traject_data = [np.flip(points) for points in traject_data]
        elif isinstance(traject_data, list):
            [points.reverse() for points in traject_data]

    return traject_data


def flip(data):
    rows = __extract_rows(__parse_data(data))
    rows_reversed = reverse(rows)
    traject_data = merge_rows(rows_reversed)

    return traject_data


def extend(data, side: str, length: float, row_number: Union[int, list[int]] = -1):
    rows = __extract_rows(__parse_data(data))
    row_numbers = __filter_row_numbers(rows, row_number)

    if side == 'begin':
        side_order = ['begin', 'end']
        for i in row_numbers:
            __extend_row(rows[i], side_order[i % 2], length)
    elif side == 'end':
        side_order = ['end', 'begin']
        for i in row_numbers:
            __extend_row(rows[i], side_order[i % 2], length)
    elif side == 'both':
        for i in row_numbers:
            __extend_row(rows[i], 'both', length)

    traject_data = merge_rows(rows)

    return traject_data


def __filter_row_numbers(rows, row_number: Union[int, list[int]] = -1):
    row_numbers = []
    if isinstance(row_number, int):
        row_numbers = range(len(rows)) if row_number < 0 else [row_number]
    if isinstance(row_number, list):
        row_numbers = row_number

    # Filter out rows that do not exist
    return [num for num in row_numbers if 0 <= num < len(rows)]


def shift_alternate(data, distance: float, row_number: Union[int, list[int]] = -1):
    rows = __extract_rows(__parse_data(data))
    row_numbers = __filter_row_numbers(rows, row_number)

    for i in row_numbers:
        __shift_row(rows[i], distance)

    traject_data = merge_rows(rows)

    return traject_data


def shift(data, distance: float, row_number: Union[int, list[int]] = -1):
    rows = __extract_rows(__parse_data(data))
    row_numbers = __filter_row_numbers(rows, row_number)

    for i in row_numbers:
        __shift_row(rows[i], distance * (-1 if i % 2 == 0 else 1))

    traject_data = merge_rows(rows)

    return traject_data


def add(data, distance: float, side: str = 'begin', number: int = 1):
    rows = __extract_rows(__parse_data(data))

    index = 0 if side == 'begin' else -1
    for i in range(number):
        if side == 'begin':
            rows = [reverse(deepcopy(rows[index]))] + rows
        else:
            rows = rows + [reverse(deepcopy(rows[index]))]
        __shift_row(rows[index], distance)

    traject_data = merge_rows(rows)

    return traject_data


def remove(data, row_number: Union[int, list[int]] = None):
    rows = __extract_rows(__parse_data(data))
    row_numbers = __filter_row_numbers(rows, row_number)

    if len(row_numbers) == len(rows):
        return data

    for x, row_number in enumerate(row_numbers):
        # Remove row
        del rows[row_number]
        # Reverse all other rows
        for i in range(row_number, len(rows)):
            rows[i] = reverse(rows[i])
        # Substract 1 from the remaining row numbers
        for i in range(x + 1, len(row_numbers)):
            row_numbers[i] = row_numbers[i] - 1

    traject_data = merge_rows(rows)

    return traject_data



import json
import redis
from os import path
from artof_utils.singleton import Singleton
from artof_utils.schemas.state import State


class RedisServer(metaclass=Singleton):
    previous_rt_dict: dict = {}

    def __init__(self, ilvo_path="", host: str = '127.0.0.1', port: int = 6379):
        self.variables = dict()

        if ilvo_path:
            config_path = path.join(ilvo_path, 'config.json')
            assert path.isfile(config_path), "'%s' is not an existing file" % config_path
            with open(config_path, mode='r') as json_file:
                self.config = json.load(json_file)

            types_path = path.join(ilvo_path, 'types.json')
            assert path.isfile(types_path), "'%s' is not an existing file" % types_path
            with open(types_path, mode='r') as json_file:
                self.types = json.load(json_file)

            self.load_variables("plc.monitor", self.config["variables"]["plc"]["monitor"])
            self.load_variables("plc.control", self.config["variables"]["plc"]["control"])
            self.load_variables("pc", self.config["variables"]["pc"])

            # A problem for the CI/CD
            # if "redis" in self.config["protocols"]:
            #     host = self.config["protocols"]["redis"]["ip"]
            #     port = self.config["protocols"]["redis"]["port"]

        self.r = redis.Redis(host=host, port=port)
        print('Connection to redis server host: %s:%d done!' % (host, port))

    def variable_in_config(self, name):
        return len(self.variables.keys()) > 0 and name in self.variables.keys()

    def load_variables(self, name: str, variables: dict):
        for key, variable_type in variables.items():
            if isinstance(variable_type, str):
                if variable_type in self.types:
                    self.load_variables(name + '.' + key, self.types[variable_type])
                else:
                    if "array" in variable_type:
                        array_type = variable_type.split(" of ")[-1]
                        array_size_str = variable_type.split(" of ")[0].split("[")[-1].replace("]", "")
                        array_size = int(array_size_str)

                        for i in range(array_size):
                            variable_name = name + '.' + key + '.' + str(i)
                            self.variables[variable_name] = {"type": array_type}
                    else:
                        variable_name = name + '.' + key
                        self.variables[variable_name] = {"type": variable_type}
            else:
                self.load_variables(name + '.' + key, variable_type)

    def convert_dtype(self, value: str, variable: str):
        """
        Convert the redis value to the prefered data type
        :param value: string of the redis value
        :param variable: string of the complete variable name
        :return: the value in the correct datatype
        """
        if value == "(nil)" or value is None:
            value = None

        var_type = "string"
        if self.variable_in_config(variable):
            var_type = self.variables[variable]["type"]

        if var_type in ["float", "double"]:
            try:
                val = float(value)
            except TypeError:
                val = 0.0
        elif 'int' in var_type:
            try:
                val = int(float(value))
            except TypeError:
                val = 0
        elif var_type in ["bool"]:
            try:
                val = bool(int(value))
            except TypeError:
                val = False
            except ValueError:
                v = False if value is None else value.decode()
                val = (v == "1" or v.lower() == "true")

        elif var_type in ["string"]:
            try:
                val = "" if value is None else value.decode()
            except TypeError:
                val = "-"
        else:
            try:
                val = value
            except TypeError:
                val = False
        return val

    def set_value(self, name, value) -> None:
        """
        Set the value of one parameter
        :return: None
        """
        var_name = name
        value_str = ('true' if value else 'false') if isinstance(value, bool) else str(value)
        if not self.r.set(var_name, value_str):
            print("Note: Setting variable \'%s\' to value \'%s\' failed" % (var_name, value_str))

    def set_n_values(self, variables: dict) -> None:
        """
        Set multile variables at the same time
        :param variables: dictionary containing 'variable' 'value' key value pairs
        :return: None
        """
        variables = {var: ('true' if val else 'false') if isinstance(val, bool) else str(val)
                     for var, val in variables.items()}
        self.r.mset(variables)

    def get_all_values(self) -> dict:
        """
        Get all variable values from the variables in the configuration file
        :return: dict
        """
        all_variables = list(self.variables.keys())
        d = self.get_n_values(all_variables)

        return d

    def get_value(self, name: str):
        var_name = name
        value_str = self.r.get(var_name)
        return self.convert_dtype(value_str, var_name)

    def get_n_values(self, variables: list[str]) -> dict:
        """
        Get multiple variables at once
        :param variables: list with all variables names of values want
        :return: dict with all variables values
        """
        values = self.r.mget(variables)
        d = {var: self.convert_dtype(val, var) for val, var in zip(values, variables)}

        return d

    def get_json_value(self, var_name: str) -> dict:
        """
        Get json object
        :param var_name: Variable name
        :return: read json object as dictionary
        """
        ret = self.r.json().get(var_name)
        if ret == "(nil)":
            return {}

        return ret

    def get_n_json_value(self, var_names: list, path_: str = '$') -> dict:
        """
        Get n json values
        :param var_names: List of variables to get
        :param path_: JSONpath to variables default is root
        :return: List of dictionaries
        """
        val = self.r.json().mget(var_names, path_)
        val_dict = {var: (v[0] if v is not None else v) for var, v in zip(var_names, val)}
        return val_dict

    def get_state(self, state_name: str):
        return State(self.get_json_value(state_name))

    def set_state(self, state_name: str, state: State):
        self.set_json_value(state_name, state.model_dump())

    def set_json_value(self, var_name: str, value, json_path='$'):
        """
        Set json object
        :param json_path: path to json value
        :param var_name: Variable name
        :param value: write json object as dictionary
        :return: None
        """
        self.r.json().set(var_name, json_path, value)



from unittest import TestCase
import numpy as np
from artof_utils.helpers.array import get_depth


class TestHelpersArray(TestCase):

    def test_get_depth_list_0(self):
        self.assertEqual(get_depth(4), 0)
        self.assertEqual(get_depth(None), 0)
        self.assertEqual(get_depth(False), 0)
        self.assertEqual(get_depth("string"), 0)

    def test_get_depth_list_empty(self):
        depth = get_depth([])
        self.assertEqual(depth, 1)

    def test_get_depth_list_1(self):
        depth = get_depth([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(depth, 1)

    def test_get_depth_np_1(self):
        depth = get_depth(np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))

        self.assertEqual(depth, 1)

    def test_get_depth_list_2(self):
        depth = get_depth([[0, 1], [0, 1], [0, 1]])

        self.assertEqual(depth, 2)

    def test_get_depth_np_2(self):
        depth = get_depth(np.array([[0, 1], [0, 1], [0, 1]]))

        self.assertEqual(depth, 2)

    def test_get_depth_list_2_special(self):
        depth = get_depth([[0, 1], 1, None, [0, 1]])

        self.assertEqual(depth, 2)

    def test_get_depth_list_3(self):
        depth = get_depth([[[0, 1], [0, 1], [0, 1]], [[0, 1], [0, 1], [0, 1]]])

        self.assertEqual(depth, 3)

    def test_get_depth_np_3(self):
        depth = get_depth(np.array([[[0, 1], [0, 1], [0, 1]], [[0, 1], [0, 1], [0, 1]]]))

        self.assertEqual(depth, 3)



from dotenv import load_dotenv
from os import path, getenv

load_dotenv()

ilvo_path = getenv('ILVO_PATH')
if ilvo_path and path.exists(ilvo_path):
    base = path.join(ilvo_path)
    platform_settings = path.join(base, 'settings.json')
    implements = path.join(base, 'implement')
    fields = path.join(base, 'field')



import time

from artof_utils.singleton import Singleton
from artof_utils.schemas.settings import load_settings
from artof_utils.schemas.field import Field, get_current_field_name
from artof_utils.schemas.hitches import Hitches
from artof_utils.schemas.navigation import Navigation
from artof_utils.schemas.state import State
from artof_utils.helpers import hardware as hw
from artof_utils.helpers import shape as shp
from artof_utils.helpers import polygon
from artof_utils.schemas.settings import AutoMode
from artof_utils.redis_instance import redis_server
from shapely.geometry import Point


# Robot Manager
class RobotManager(metaclass=Singleton):
    """
    A singleton class responsible for managing the robot's settings, field information,
    navigation, and interaction with external resources like redis_server. This ensures a single
    instance can manage and maintain the state and behavior of the robotic platform.
    """

    def __init__(self):
        """
        Initializes the RobotManager instance by loading platform settings and field
        information, setting up hitches and navigation objects, and preparing state
        and voltage variables.
        """
        self.platform_settings = None
        self.field = None

        # Load settings and field from configuration files
        self.load_settings()
        self.load_field()

        # Initialize Hitches and Navigation objects
        self.hitches = Hitches()
        self.navigation = Navigation()

        # Configure platform hitches based on loaded settings
        if self.platform_settings:
            self.hitches.add_hitches(self.platform_settings.hitches)

        # Retrieve robot state and voltage information from Redis
        all_keys = redis_server.variables.keys()
        self.robot_state_vars = [key for key in all_keys if key.startswith('plc.monitor.state.')]

    def load_settings(self):
        """Loads platform settings from a JSON file specified in the paths configuration."""
        print("Load settings")
        self.platform_settings = load_settings()

    def load_field(self):
        """Loads the current field configuration based on the field name."""
        print("Load Field")
        self.field = Field(get_current_field_name())

    # Getter and setters
    @staticmethod
    def get_navigation_modes():
        return [(mode.id, mode.name) for mode in robot_manager.platform_settings.nav_modes]

    def get_navigation_states(self):
        """
        Retrieves a list of available navigation states for the robot, including
        any custom auto modes defined in the platform settings.

        Returns:
            list: A list of navigation state names.
        """
        auto_modes_settings = [AutoMode.model_validate({'name': 'normal', 'id': 0})] + self.platform_settings.auto_modes
        state_names = [auto_mode_setting.name for auto_mode_setting in auto_modes_settings]
        return state_names

    def get_navigation_state(self):
        state_vars = redis_server.get_n_values(self.robot_state_vars)
        active_states = [k.replace('plc.monitor.state.', '') for k, v in state_vars.items() if v]
        current_state = active_states[0] if len(active_states) > 0 else robot_manager.get_navigation_states()[0]

        if self.get_simulation_mode() and self.get_simulation_auto():
            current_state = 'auto'

        return current_state

    def set_navigation_state(self, navigation_state):
        """
        Sets the robot's navigation state based on the provided state name.

        Parameters:
            navigation_state (str): The name of the navigation state to set.
        """
        sim_mode = self.get_simulation_mode()
        if sim_mode:
            redis_server.set_value('pc.simulation.auto', navigation_state != 'normal')
        else:
            if navigation_state in self.get_navigation_states():
                redis_server.set_value('plc.monitor.state.' + navigation_state, True)

        # Control operations
        control_name = 'plc.control.state.' + navigation_state
        if navigation_state in self.get_navigation_states() and control_name in redis_server.variables.keys():
            # Pulse the state variable to trigger state change
            redis_server.set_value(control_name, True)
            time.sleep(0.5)
            redis_server.set_value(control_name, False)
            print("Pulsed %s" % control_name)

    def set_position_latlon(self, lat, lon):
        """
        Set the position of the robot based on latitude and longitude coordinates.

        :param lat: latitude coordinate
        :param lon: longitude coordinate
        :return: None
        """
        # Define the coordinate reference systems
        wgs84_crs = 'EPSG:4326'  # WGS 84
        utm_crs = 'EPSG:%d' % self.field.shp_geofence.gdf.crs.to_epsg()

        x, y = shp.transform_crs(wgs84_crs, utm_crs, [lat, lon])
        self.set_position(x, y)

    @staticmethod
    def set_position(x, y, yaw=None):
        robot_ref_state = redis_server.get_json_value("robot.ref.state")
        if robot_ref_state is None:
            # skip if there is no robot_ref_state
            return
        robot_ref_state["T"] = [x, y, 0.0]
        if yaw is not None:
            robot_ref_state["R"] = [0.0, 0.0, yaw]
        redis_server.set_json_value("robot.ref.state", robot_ref_state)

    @staticmethod
    def set_velocity(vx, omega):
        redis_server.set_value('plc.control.navigation.velocity.longitudinal', vx)
        redis_server.set_value('plc.control.navigation.velocity.angular', omega)

    @staticmethod
    def get_velocity():
        vx = redis_server.get_value('plc.control.navigation.velocity.longitudinal')
        omega = redis_server.get_value('plc.control.navigation.velocity.angular')

        return vx, omega

    @staticmethod
    def get_simulation_mode():
        """
        Get the current simulation mode from the Redis server.

        :return: The current simulation mode (str)
        """
        return redis_server.get_value('pc.simulation.active')

    @staticmethod
    def get_simulation_auto():
        return redis_server.get_value('pc.simulation.auto')

    def set_simulation_mode(self, active=True):
        redis_server.set_value('pc.simulation.active', active)
        # set position to first point of the traject
        if len(self.field.shp_traject.gdf):
            traject_points = robot_manager.field.shp_traject.gdf.geometry[0].coords

            first_point = Point(traject_points[0])
            second_point = Point(traject_points[1])
            path_orientation = shp.get_orientation(first_point, second_point)

            self.set_position(first_point.x, first_point.y, path_orientation)
        # navigation state back to normal when simulation mode is turned off
        if not active:
            self.set_navigation_state('normal')

    @staticmethod
    def set_simulation_speed_factor(factor):
        redis_server.set_value('pc.simulation.factor', factor)

    @staticmethod
    def get_simulation_speed_factor():
        return redis_server.get_value('pc.simulation.factor')

    @staticmethod
    def get_programming_mode():
        return redis_server.get_value('plc.monitor.substate.programming')

    @staticmethod
    def acknowledge_notification():
        redis_server.set_value('pc.execution.notification', '-')

    @staticmethod
    def update_field():
        redis_server.set_value('pc.field.updated', True)

    def status(self):
        return redis_server.get_json_value("robot.status")

    def context(self):
        redis_data = redis_server.get_n_json_value(["robot.center.state", "robot.ref.state", "robot.head.state",
                                                    "robot.contour", "hitch.states", "implement.states",
                                                    "navigation.controller.info"])
        r = dict()

        r['robot'] = {'contours': redis_data['robot.contour'],
                      'center': redis_data['robot.center.state']['point'],
                      'ref': redis_data['robot.ref.state']['point'],
                      'head': redis_data['robot.head.state']['point'],
                      'orientation': redis_data['navigation.controller.info']['heading']}
        # Add each hitch
        r['hitches'] = redis_data['hitch.states']

        # Add implements
        r['implements'] = redis_data['implement.states']

        # Add controller info
        r['controller_info'] = redis_data['navigation.controller.info']

        return r


robot_manager = RobotManager()



from artof_utils.redis import RedisServer
import artof_utils.paths as paths
import ipaddress
from os import getenv


# Redis server
def __is_numerical(input_str):
    try:
        # Try converting the string to a float
        float_value = float(input_str)
        return True
    except ValueError:
        # If ValueError is raised, it means the conversion failed
        return False


def __is_valid_ip(ip_str):
    try:
        ipaddress.ip_address(ip_str)
        return True
    except ValueError:
        return False


redis_server = RedisServer(ilvo_path=paths.ilvo_path,
                           host=getenv('REDIS_HOST') if getenv('REDIS_HOST') else '127.0.0.1',
                           port=int(getenv('REDIS_PORT'))
                           if getenv('REDIS_PORT') and __is_numerical(getenv('REDIS_PORT'))
                           else 6379)



import matplotlib.pyplot as plt
import numpy as np


def plot_on_top(paths: dict):
    # Plot result
    lw = 1

    # Draw orientation vectors
    fig, ax = plt.subplots(figsize=(10, 5))
    colors = ['red', 'orange', 'green', 'blue']
    for num, (label, coords) in enumerate(paths.items()):
        color = colors[num % len(colors)]
        np_path = np.array(coords)
        x_path = np_path[:, 0]
        y_path = np_path[:, 1]
        ax.scatter(x_path, y_path, color=color, marker=".", label=label)
        ax.plot(x_path, y_path, color=color, linestyle='dashed', linewidth=lw)
        for i, txt in enumerate(range(len(x_path))):
            ax.annotate("%d" % i, (float(x_path[i]), float(y_path[i])),
                        textcoords="offset points", xytext=(0, 10), ha='center')

        ax.legend(loc="lower left")

    plt.show()


def plot_side_by_side(paths: dict):
    # Plot result
    lw = 1

    # Draw orientation vectors
    fig, axs = plt.subplots(1, len(paths.keys()), figsize=(10, 5))

    num = 0
    for num, (title, coords) in enumerate(paths.items()):
        np_path = np.array(coords)
        x_path = np_path[:, 0]
        y_path = np_path[:, 1]
        axs[num].scatter(x_path, y_path, color='green', marker=".")
        axs[num].plot(x_path, y_path, color='green', linestyle='dashed', linewidth=lw)
        for i, txt in enumerate(range(len(x_path))):
            axs[num].annotate(txt, (float(x_path[i]), float(y_path[i])),
                              textcoords="offset points", xytext=(0, 10), ha='center')

        axs[num].set_title(title)

        num += 1

    plt.show()



from pydantic import BaseModel
from typing import Optional
from artof_utils.redis_instance import redis_server


class Navigation(BaseModel):
    navigation_mode: int
    non_operational_velocity: float
    operational_velocity: float
    carrot_distance: Optional[float]
    weight_factor: Optional[float] = 0.0
    kp_steady_state: Optional[float] = 0.0
    ki_steady_state: Optional[float] = 0.0
    kd_steady_state: Optional[float] = 0.0
    kp_rough: Optional[float] = 0.0
    ki_rough: Optional[float] = 0.0
    kd_rough: Optional[float] = 0.0

    redis_variables: list[str]

    def __init__(self):
        redis_variables_ = [
            "pc.navigation.mode",
            "pc.navigation.non_operational_velocity",
            "pc.navigation.operational_velocity",
            "pc.purepursuit.weight_factor",
            "pc.purepursuit.carrot_distance",
            "pc.pid_steady_state.p",
            "pc.pid_steady_state.i",
            "pc.pid_steady_state.d",
            "pc.pid_rough.p",
            "pc.pid_rough.i",
            "pc.pid_rough.d"
        ]
        d = redis_server.get_n_values(redis_variables_)

        super().__init__(navigation_mode=d['pc.navigation.mode'],
                         non_operational_velocity=d['pc.navigation.non_operational_velocity'],
                         operational_velocity=d['pc.navigation.operational_velocity'],
                         weight_factor=d['pc.purepursuit.weight_factor'],
                         carrot_distance=d['pc.purepursuit.carrot_distance'],
                         kp_steady_state=d['pc.pid_steady_state.p'],
                         ki_steady_state=d['pc.pid_steady_state.i'],
                         kd_steady_state=d['pc.pid_steady_state.d'],
                         kp_rough=d['pc.pid_rough.p'],
                         ki_rough=d['pc.pid_rough.i'],
                         kd_rough=d['pc.pid_rough.d'],
                         redis_variables=redis_variables_)

    def update(self):
        d = redis_server.get_n_values(self.redis_variables)
        self.navigation_mode = d['pc.navigation.mode']
        self.non_operational_velocity = d['pc.navigation.non_operational_velocity']
        self.operational_velocity = d['pc.navigation.operational_velocity']
        self.weight_factor = d['pc.purepursuit.weight_factor']
        self.kp_steady_state = d['pc.pid_steady_state.p']
        self.ki_steady_state = d['pc.pid_steady_state.i']
        self.kd_steady_state = d['pc.pid_steady_state.d']
        self.kp_rough = d['pc.pid_rough.p']
        self.ki_rough = d['pc.pid_rough.i']
        self.kd_rough = d['pc.pid_rough.d']
        self.carrot_distance = d['pc.purepursuit.carrot_distance']

    @staticmethod
    def change(navigation_mode: int, non_operational_velocity: float, operational_velocity: float,
               carrot_distance: float, weight_factor: float,
               kp_steady_state: float, ki_steady_state: float, kd_steady_state: float,
               kp_rough: float, ki_rough: float, kd_rough: float):
        d = {
            'pc.navigation.mode': navigation_mode,
            'pc.navigation.non_operational_velocity': non_operational_velocity,
            'pc.navigation.operational_velocity': operational_velocity,
            'pc.purepursuit.weight_factor': weight_factor,
            'pc.pid_steady_state.p': kp_steady_state,
            'pc.pid_steady_state.i': ki_steady_state,
            'pc.pid_steady_state.d': kd_steady_state,
            'pc.pid_rough.p': kp_rough,
            'pc.pid_rough.i': ki_rough,
            'pc.pid_rough.d': kd_rough,
            'pc.purepursuit.carrot_distance': carrot_distance
        }
        redis_server.set_n_values(d)

    @property
    def context(self):
        return self.model_dump(exclude={'redis_variables'})



from pydantic import BaseModel


class SimStatus(BaseModel):
    active_sim: bool
    sim_active_auto: bool


class KeyboardCommand(BaseModel):
    vel_fwd: float
    omega: float


class SimPosition(BaseModel):
    x: float
    y: float
    yaw: float


class SpeedFactor(BaseModel):
    factor: float



from unittest import TestCase
from artof_utils.schemas.state import State
from artof_utils.redis import RedisServer
from os import getenv
import artof_utils.paths as paths


class TestRedisServer(TestCase):
    def test_set_get_value_from_config(self):
        variables = {
            "pc.gps.fix": 5,
            "pc.gps.hrp_mode": 3,
        }

        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        for k, v in variables.items():
            self.assertTrue(redis_server.variable_in_config(k))
            redis_server.set_value(k, v)

        for k, v in variables.items():
            value = redis_server.get_value(k)
            self.assertEqual(v, value)

    def test_n_set_get_value_from_config(self):
        variables = {
            "random1": 15.0,
            "random2": 15.0,
        }
        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_n_values(variables)

        r_variables = redis_server.get_n_values(list(variables.keys()))
        for k, v in r_variables.items():
            self.assertEqual(str(v), str(variables[k]))

    def test_n_set_get_value_partly_config(self):
        variables = {
            "pc.gps.fix": 5,
            "pc.gps.hrp_mode": 3,
            "random": 10.0
        }
        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_n_values(variables)

        r_variables = redis_server.get_n_values(list(variables.keys()))
        for k, v in r_variables.items():
            self.assertEqual(str(v), str(variables[k]))

    def test_n_set_get_value_no_config(self):
        variables = {
            "random0": 20.0,
            "random1": 2.0,
            "random2": 4.0
        }
        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_n_values(variables)

        r_variables = redis_server.get_n_values(list(variables.keys()))
        for k, v in r_variables.items():
            self.assertEqual(str(v), str(variables[k]))

    def test_set_get_json_value(self):
        j_orig = {
            "int": 1.0,
            "double": 10.0,
            "str": "str"
        }
        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_json_value("json_obj", j_orig)

        j_read = redis_server.get_json_value("json_obj")
        self.assertEqual(str(j_read), str(j_orig))

    def test_set_get_json_state(self):
        d = {
            "R": [-0.0, 0.0, 40.0],
            "R_cov": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
            "T": [554579.00, 5648161.00, 0.0],
            "T_cov": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]}
        state = State(d)
        state_name = "robot.ref.state"

        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_state(state_name, state)
        state_ = redis_server.get_state(state_name)

        print(f"state: {state}")
        print(f"state_: {state_}")
        assert state == state_, f"States are not equal!"

    def test_json_value_with_path(self):
        d = {
            "test": [
                {"nested_1": {
                    "json": 1
                }},
                {"nested_2": {
                    "json": 1
                }}
            ]
        }
        json_name = "test"

        redis_server = RedisServer(ilvo_path=paths.ilvo_path, host='127.0.0.1')
        redis_server.set_json_value(json_name, d)

        redis_server.set_json_value(json_name, 0, '$.test[0].nested_1.json')
        json_value_test = redis_server.get_json_value(json_name)

        print(f"json_value \'test\' before set: ", d)
        print(f"json_value \'test\' after set: ", json_value_test)
        # assert state == state_, f"States are not equal!"

    def test_load_variables(self):
        # Arrange
        # Act
        redis_server = RedisServer(ilvo_path=paths.ilvo_path)
        # Assert
        self.assertGreaterEqual(len(redis_server.variables), 0)



from unittest import TestCase
from artof_utils.shapefile import Shapefile, GeomType
from os import path
import numpy as np


class TestShapefile(TestCase):
    def test_init_empty_shapefile(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'empty_new')
        # Act
        shapefile = Shapefile(folder_path)
        # Assert
        self.assertTrue(path.exists(shapefile.file_path))

    def test_context_linestring(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'linestring')
        shapefile = Shapefile(folder_path)
        ref_points = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0]]])
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)
        self.assertTrue(np.array(context['paths']).shape == np.array(context['latlng']).shape)
        self.assertTrue(np.allclose(ref_points, np.array(context['paths'])))

    def test_context_polygon(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygon')
        shapefile = Shapefile(folder_path)
        ref_points = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0], [0.0, 0.0]]])
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)
        self.assertTrue(np.array(context['rings']).shape == np.array(context['latlng']).shape)
        self.assertTrue(np.allclose(ref_points, np.array(context['rings'][0])))

    def test_context_polygons(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygons')
        shapefile = Shapefile(folder_path)
        ref_points = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0], [0.0, 0.0]]])
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)
        self.assertTrue(np.allclose(ref_points, np.array(context['rings'][0])))
        self.assertEqual(len(context['rings']), 3)
        self.assertEqual(len(context['latlng']), 3)

    def test_context_points(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'points')
        shapefile = Shapefile(folder_path)
        ref_points = np.array([[0.0, 0.0], [100.0, 0.0], [100.0, 100.0], [0.0, 100.00]])
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)
        self.assertTrue(np.allclose(ref_points, np.array(context['points'])))
        self.assertTrue(np.array(context['points']).shape == np.array(context['latlng']).shape)

    def test_context_multipoints(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'multipoints')
        shapefile = Shapefile(folder_path)
        ref_points = np.array([[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0]])
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)
        self.assertTrue(np.allclose(ref_points, np.array(context['points'])))
        self.assertTrue(np.array(context['points']).shape == np.array(context['latlng']).shape)

    def test_context_multipoints2(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'multipoints2')
        shapefile = Shapefile(folder_path)
        # Act
        context = shapefile.context
        # Assert
        self.assertEqual(context['wkid'], 32631)
        self.assertEqual(context['hasZ'], False)

    def test_new_polygon(self):
        # Arrange
        coords_1 = [[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0]]]
        coords_2 = np.array([[0.0, 0.0], [0.0, 100.00], [100.0, 100.0]])
        coords_3 = [[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0], [0.0, 0.0]]]
        coords_4 = [coords_1[0], coords_2, coords_3[0]]
        coords_1_ref = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0], [0.0, 0.0]]])
        coords_2_ref = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [0.0, 0.0]]])
        coords_3_ref = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0], [0.0, 0.0]]])
        coords_4_ref = [coords_3_ref[0], coords_2_ref[0], coords_3_ref[0]]
        folder_path_1 = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygon1_new')
        folder_path_2 = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygon2_new')
        folder_path_3 = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygon3_new')
        folder_path_4 = path.join(path.dirname(__file__), 'files', 'shapefile', 'polygon4_new')
        # Act
        shapefile_1 = Shapefile(folder_path_1)
        shapefile_2 = Shapefile(folder_path_2)
        shapefile_3 = Shapefile(folder_path_3)
        shapefile_4 = Shapefile(folder_path_4)
        shapefile_1.update(coords_1, GeomType.POLYGON)
        shapefile_2.update(coords_2, GeomType.POLYGON)
        shapefile_3.update(coords_3, GeomType.POLYGON)
        shapefile_4.update(coords_4, GeomType.POLYGON)
        # Assert
        self.assertEqual(shapefile_1.context['wkid'], 32631)
        self.assertEqual(shapefile_2.context['wkid'], 32631)
        self.assertEqual(shapefile_3.context['wkid'], 32631)
        self.assertEqual(shapefile_4.context['wkid'], 32631)
        self.assertTrue(path.exists(shapefile_1.file_path))
        self.assertTrue(path.exists(shapefile_2.file_path))
        self.assertTrue(path.exists(shapefile_3.file_path))
        self.assertTrue(path.exists(shapefile_4.file_path))
        self.assertTrue(np.allclose(coords_1_ref, np.array(shapefile_1.context['rings'])))
        self.assertTrue(np.allclose(coords_2_ref, np.array(shapefile_2.context['rings'])))
        self.assertTrue(np.allclose(coords_3_ref, np.array(shapefile_3.context['rings'])))
        self.assertTrue(np.array(shapefile_1.context['rings']).shape == np.array(shapefile_1.context['latlng']).shape)
        self.assertTrue(np.array(shapefile_2.context['rings']).shape == np.array(shapefile_2.context['latlng']).shape)
        self.assertTrue(np.array(shapefile_3.context['rings']).shape == np.array(shapefile_3.context['latlng']).shape)

        for i in range(len(coords_4_ref)):
            self.assertTrue(np.allclose(coords_4_ref[i], np.array(shapefile_4.context['rings'][i])))

    def test_new_linestring(self):
        # Arrange
        coords_1 = np.array([[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0]]])
        coords_2 = [[0.0, 0.0], [0.0, 100.00], [100.0, 100.0]]
        folder_path_1 = path.join(path.dirname(__file__), 'files', 'shapefile', 'linestring1_new')
        folder_path_2 = path.join(path.dirname(__file__), 'files', 'shapefile', 'linestring2_new')
        coords_1_ref = coords_1
        coords_2_ref = coords_1
        # Act
        shapefile_1 = Shapefile(folder_path_1)
        shapefile_2 = Shapefile(folder_path_2)
        shapefile_1.update(coords_1, GeomType.LINESTRING)
        shapefile_2.update(coords_2, GeomType.LINESTRING)
        # Assert
        self.assertEqual(shapefile_1.context['wkid'], 32631)
        self.assertEqual(shapefile_2.context['wkid'], 32631)
        self.assertTrue(path.exists(shapefile_1.file_path))
        self.assertTrue(path.exists(shapefile_2.file_path))
        self.assertTrue(np.allclose(coords_1_ref, np.array(shapefile_1.context['paths'])))
        self.assertTrue(np.allclose(coords_2_ref, np.array(shapefile_2.context['paths'])))
        self.assertTrue(np.array(shapefile_1.context['paths']).shape == np.array(shapefile_1.context['latlng']).shape)
        self.assertTrue(np.array(shapefile_2.context['paths']).shape == np.array(shapefile_2.context['latlng']).shape)

    def test_new_points(self):
        # Arrange
        coords_1 = [[0.0, 0.0], [0.0, 100.00], [100.0, 100.0]]
        folder_path_1 = path.join(path.dirname(__file__), 'files', 'shapefile', 'points1_new')
        # Act
        shapefile_1 = Shapefile(folder_path_1)
        shapefile_1.update(coords_1, GeomType.POINT)
        # Assert
        self.assertEqual(shapefile_1.context['wkid'], 32631)
        self.assertTrue(path.exists(shapefile_1.file_path))
        self.assertTrue(np.allclose(coords_1, np.array(shapefile_1.context['points'])))
        self.assertTrue(np.array(shapefile_1.context['points']).shape == np.array(shapefile_1.context['latlng']).shape)

    def test_shapefile_empty(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'empty')
        # Act
        shapefile = Shapefile(folder_path)
        # Assert
        self.assertEqual(shapefile.context['empty'], True)

    def test_update_latlng(self):
        # Arrange
        folder_path = path.join(path.dirname(__file__), 'files', 'shapefile', 'traject_latlng_new')
        shapefile = Shapefile(folder_path)
        data_utm = [[[554356.9501732234, 5647897.73942658], [554304.8298267765, 5647995.610573417],
                     [554306.4898267768, 5647996.4905734155], [554358.6101732232, 5647898.619426582],
                     [554359.8901732232, 5647899.289426582], [554307.7698267768, 5647997.160573415],
                     [554309.0898267763, 5647997.870573415], [554361.2101732236, 5647899.999426585],
                     [554362.5401732236, 5647900.699426582], [554310.4198267765, 5647998.570573414],
                     [554311.7398267763, 5647999.280573415], [554363.8601732237, 5647901.409426584],
                     [554365.1801732233, 5647902.109426583], [554313.0598267767, 5647999.980573417],
                     [554314.3898267763, 5648000.690573414], [554366.5101732236, 5647902.819426582],
                     [554367.830173223, 5647903.519426583], [554315.7098267768, 5648001.390573415],
                     [554317.0398267768, 5648002.100573416], [554369.1601732231, 5647904.229426583],
                     [554370.4801732232, 5647904.9294265825], [554318.3598267768, 5648002.800573416]]]
        data_latlng = [[[50.9801038330035, 3.7743247519946936], [50.98098878064322, 3.7735969850426625],
                        [50.98099653675816, 3.7736207616938873], [50.98011158897294, 3.7743485283119287],
                        [50.980117492559124, 3.774366860583644], [50.98100244045656, 3.773639094224294],
                        [50.981008700046935, 3.773658002499855], [50.98012375203382, 3.7743857685950313],
                        [50.980129920644025, 3.7744048175536555], [50.98101486877373, 3.77367705172676],
                        [50.981021128357916, 3.7736959600124296], [50.98013618011252, 3.774423725575141],
                        [50.98014234966079, 3.7744426321057896], [50.98102729802194, 3.7737148668086364],
                        [50.981033556656456, 3.7737339175451745], [50.98014860817873, 3.7744616825754176],
                        [50.98015477772084, 3.7744805891160937], [50.981039726314265, 3.7737528243514107],
                        [50.98104598494257, 3.7737718750980958], [50.98016103623252, 3.7744996395958577],
                        [50.98016720576842, 3.774518546146571], [50.981052154594174, 3.7737907819143555]]]
        # Act
        shapefile.update(data_latlng, GeomType.LINESTRING, epsg=4326)
        # Assert
        self.assertTrue(np.allclose(np.array(data_utm), list(shapefile.gdf.geometry[0].coords)))



from artof_utils.helpers import traject
from tests.visual_support import plot_side_by_side
import pandas as pd

file_name = "files/traject/simple.csv"

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    df = pd.read_csv(file_name)
    df = df.apply(pd.to_numeric, errors='coerce')

    # Calculate orientation of points
    path = [[xy[0], xy[1]] for xy in zip(df['X'], df['Y'])]
    path_reverse = traject.perform(traject.Operation.REVERSE, path)
    path_flip = traject.perform(traject.Operation.FLIP, path)

    plot_side_by_side({"Original": path, "Reversed": path_reverse, "Flipped": path_flip})



from artof_utils.singleton import Singleton
from artof_utils.schemas.implement import Implement
from glob import glob
from os import path, remove
import artof_utils.paths as paths
import json


class ImplementManager(metaclass=Singleton):
    @property
    def implements(self):
        return [Implement.load(path.basename(implement_name).replace('.json', ''))
                for implement_name in glob(path.join(paths.implements, '*.json'))]

    @staticmethod
    def update_implement(name, implement: Implement):
        implement_file_path = path.join(paths.implements, '%s.json' % name)
        with open(implement_file_path, 'w') as json_file:
            json.dump(implement.context, json_file, indent=4)

    @staticmethod
    def add_implement(name, implement: Implement):
        ImplementManager.update_implement(name, implement)

    @staticmethod
    def remove_implement(name):
        implement_file_path = path.join(paths.implements, '%s.json' % name)
        if path.exists(implement_file_path):
            remove(implement_file_path)

    @staticmethod
    def get_implement(name):
        return Implement.load(name)


implement_manager = ImplementManager()



from unittest import TestCase

import numpy as np

from artof_utils.schemas.state import State
import artof_utils.paths as paths
import json


class TestRobotManager(TestCase):
    def test_load_settings(self):
        # Arrange
        # - Load settings file
        with open(paths.platform_settings, 'r') as f:
            ref_settings = json.load(f)
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager

        # Act
        # - Load settings with robot manager
        settings = robot_manager.platform_settings.model_dump()

        # Assert
        self.assertEqual(ref_settings.keys(), settings.keys())

    def test_load_field(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager

        # Act
        field_name = robot_manager.field.name

        # Assert
        self.assertIsNotNone(field_name)

    def test_load_hitches(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager

        # Act
        no_hitches = len(robot_manager.hitches.hitches)

        # Assert
        self.assertGreater(no_hitches, 0)

    # def test_load_navigation(self):
    #     # Arrange
    #     # - Load settings done as singleton
    #     from artof_utils.manager import robot_manager
    #
    #     # Act
    #     nav_mode = robot_manager.navigation.navigation_mode
    #
    #     # Assert
    #     self.assertGreater(nav_mode, 0)

    # def test_shape(self):
    #     # Arrange
    #     # - Load settings done as singleton
    #     from artof_utils.robot import robot_manager
    #     state = State({'T': [10, 10, 0], 'R': [0, 0, 180.0], 'T_cov': [0, 0, 0], 'R_cov': [0, 0, 0]})
    #
    #     # Act
    #     print(robot_manager.context())

    def test_set_simulation_mode(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager
        from artof_utils.redis_instance import redis_server
        # Act
        robot_manager.set_position(0, 0)
        robot_ref_state_zero = redis_server.get_json_value("robot.ref.state")
        self.assertTrue(np.allclose(np.array(robot_ref_state_zero["T"][:2]), 0))

        robot_manager.set_simulation_mode(True)
        robot_ref_state = redis_server.get_json_value("robot.ref.state")

        first_traject_point = np.array(list(robot_manager.field.shp_traject.gdf.geometry[0].coords)[0])
        # Assert
        self.assertTrue(np.allclose(np.array(robot_ref_state["T"][:2]), first_traject_point))

    def test_stop_simulation_mode(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager
        from artof_utils.redis_instance import redis_server
        # Act
        robot_manager.set_simulation_mode(False)
        # Assert
        self.assertEqual(redis_server.get_value('pc.simulation.active'), 0)

    def test_set_simulation_speed_factor(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.robot import robot_manager
        from artof_utils.redis_instance import redis_server
        # Act
        robot_manager.set_simulation_speed_factor(0.5)
        # Assert
        self.assertEqual(redis_server.get_value('pc.simulation.factor'), 0.5)

    # def test_status(self):
    #     # Arrange
    #     # - Load settings done as singleton
    #     from artof_utils.robot import robot_manager
    #     # Act
    #     status = robot_manager.status()
    #     # Assert
    #     self.assertIsNotNone(status)
    #     self.assertTrue('fix' in status)
    #     self.assertTrue('power_level' in status)


from unittest import TestCase
import artof_utils.helpers.task as task


class TestState(TestCase):

    def test_get_hitch_choices(self):
        expected_hitches = {'FF', 'FB', 'RB'}
        hitch_choices = set([choice[0] for choice in task.get_hitch_choices()])
        self.assertEqual(hitch_choices, expected_hitches)

    def test_get_type_choices(self):
        expected_type_choices = {'hitch', 'continuous', 'discrete', 'cardan', 'intermittent'}
        type_choices = set([choice[0] for choice in task.get_type_choices()])
        self.assertEqual(type_choices, expected_type_choices)

    def test_get_implement_choices(self):
        expected_implement_choices = {'none', 'penetrometer', 'active-tool', 'auto-label', 'brander', 'spray-boom'}
        implement_choices = set([choice[0] for choice in task.get_implement_choices()])
        self.assertEqual(implement_choices, expected_implement_choices)



from unittest import TestCase
import json
import numpy as np
import math
from os import path
from artof_utils.schemas.affine import Affine


def verify_ry(R):
    angleY = R[1]
    m = np.array([
        [math.cos((math.pi / 180.0) * angleY), 0, math.sin((math.pi / 180.0) * angleY)],
        [0, 1, 0],
        [-math.sin((math.pi / 180.0) * angleY), 0, math.cos((math.pi / 180.0) * angleY)]
    ])
    return m


def verify_rz(R):
    angleZ = R[2]
    m = np.array([
        [math.cos((math.pi / 180.0) * angleZ), -math.sin((math.pi / 180.0) * angleZ), 0],
        [math.sin((math.pi / 180.0) * angleZ), math.cos((math.pi / 180.0) * angleZ), 0],
        [0, 0, 1]
    ])
    return m


def verify_vector_to_affine(t, r):
    r_ = np.matmul(verify_rz(r), verify_ry(r))
    m = np.identity(4)
    m[0:3, 0:3] = r_
    m[0:3, 3] = np.array(t).transpose()
    return m


class TestState(TestCase):
    def test_vecs_to_affine(self):
        with open(path.join(path.dirname(__file__), 'files', 'states', 'states.json')) as f:
            j_states = json.load(f)

        for j_state in j_states:
            t = np.array(j_state['state']['T'])
            r = np.array(j_state['state']['R'])

            t_, r_ = Affine().from_vectors(t, r).to_vectors()

            # If inverse angle for yaw also ok
            if abs(r_[2] - r[2]) > 0.1:
                if abs((r[2] - r_[2]) % 360.0) < 0.1:
                    r_[2] = r[2]

            print(f"Original vectors - t: {t}, r: {r}")
            print(f"Calculated vectors - t: {t_}, r: {r_}")
            print("--------------")

            self.assertTrue(np.allclose(t, t_, atol=0.1), f"Translations are not equal! "
                                                          f"Originally: {list(t)}, now: {list(t_)}")
            self.assertTrue(np.allclose(r, r_, atol=0.1), f"Rotations are not equal! "
                                                          f"Originally: {list(r)}, now: {list(r_)}")

        print("DONE!")

    def test_verify_vecs_to_affine(self):
        with open(path.join(path.dirname(__file__), 'files', 'states', 'states.json')) as f:
            j_states = json.load(f)

        for j_state in j_states:
            t = np.array(j_state['state']['T'])
            r = np.array(j_state['state']['R'])

            a_verify = verify_vector_to_affine(t, r)
            a = Affine().from_vectors(t, r).get()

            print(f"Program affine:\n {a}")
            print(f"Verify affine:\n {a_verify}")
            print("--------------")

            self.assertTrue(np.allclose(a, a_verify, atol=0.1), f"Affines are not equal! "
                                                                f"Program:\n {a} \n verify:\n {a_verify}")

        print("DONE!")



from enum import Enum

from pydantic import BaseModel, model_validator
from typing import Optional, Any
import artof_utils.paths as paths
import json


class RobotAutoMode(str, Enum):
    NORMAL = "normal"
    FULL = "auto_full"
    STEER = "auto_steer"
    THROTTLE = "auto_throttle"
    PLAYBACK = "teachandplay_play"
    END = "end_reached"


class DirNames(str, Enum):
    TRAJECT = "traject"
    GEOFENCE = "geofence"
    TEACH_AND_PLAY = "teach_and_play"
    TASK1 = "Task1"
    TASK2 = "Task2"
    TASK3 = "Task3"
    TASK4 = "Task4"
    TASK5 = "Task5"


class HitchType(str, Enum):
    HITCH = "hitch"
    CONTINUOUS = "continuous"
    DISCRETE = "discrete"
    CARDAN = "cardan"
    INTERMITTENT = "intermittent"


class HitchName(str, Enum):
    HITCH_FF = "FF"
    HITCH_FB = "FB"
    HITCH_RB = "RB"


class Transform(BaseModel):
    T: list[float] = [0.0, 0.0, 0.0]
    R: list[float] = [0.0, 0.0, 0.0]


class Hitch(BaseModel):
    id: int = 0.0
    name: HitchName = HitchName.HITCH_FB
    min: float = 0.0
    max: float = 0.0
    types: list[HitchType] = []
    transform: Transform
    float: bool = False
    # Redis variables
    setpoint: int = 0
    activate: bool = False
    start_discr_impl: bool = False
    activate_cont_impl: bool = False
    activate_cardan: bool = False
    activate_section: list[bool] = [False] * 32


class Robot(BaseModel):
    width: float = 1.0
    length: float = 1.0
    wheel_diameter: float = 0.5
    transform: Transform
    transform_center: Transform
    transform_head: Transform

    @model_validator(mode='before')
    @classmethod
    def set_transform_center(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if 'transform_center' not in data:
                data['transform_center'] = data['transform']
            if 'transform_head' not in data:
                data['transform_head'] = data['transform']
        return data


class AutoVelocity(BaseModel):
    min: float = 0.0
    max: float = 1.0


class NavMode(BaseModel):
    id: int
    name: str


class AutoMode(BaseModel):
    id: int
    name: str


class Gps(BaseModel):
    device: str
    utm_zone: int
    udp_port: Optional[int] = None
    ip: Optional[str] = None
    usb_port: Optional[str]
    ntrip_server: Optional[str]
    ntrip_mountpoint: Optional[str]
    ntrip_uname: Optional[str]
    ntrip_pwd: Optional[str]
    transform: Transform


class PlatformSettings(BaseModel):
    name: str
    robot: Robot
    auto_velocity: AutoVelocity
    nav_modes: list[NavMode]
    auto_modes: list[AutoMode]
    hitches: list[Hitch]
    gps: Gps

    @property
    def context(self):
        return self.model_dump()


def load_settings() -> PlatformSettings:
    with open(paths.platform_settings, 'r') as f:
        settings = PlatformSettings.model_validate(json.load(f))
    return settings



import numpy as np
from pyproj import Transformer
from shapely.geometry import Point
import math
from artof_utils.helpers import array


# CRS
def transform_crs(input_crs: str, output_crs: str, coordinates):
    depth = array.get_depth(coordinates)

    if input_crs == output_crs:
        return coordinates

    # Recursively calculate the transformed matrices
    if depth > 1:
        return [transform_crs(input_crs, output_crs, coords) for coords in coordinates]
    else:
        # Create a transformer from the input CRS to the output CRS
        transformer = Transformer.from_crs(input_crs, output_crs)
        # Process single coordinate
        x, y = coordinates
        return list(transformer.transform(x, y))


# Shapes
def get_orientation(point1, point2, degrees=True) -> float:
    # Calculate the difference in x and y coordinates
    if isinstance(point1, Point) and isinstance(point2, Point):
        delta_x = point2.x - point1.x
        delta_y = point2.y - point1.y
    elif isinstance(point1, list) and isinstance(point2, list):
        delta_x = point2[0] - point1[0]
        delta_y = point2[1] - point1[1]
    else:
        return 0

    # Calculate the angle (in radians) between the line segment and the x-axis
    angle_rad = math.atan2(delta_y, delta_x)

    # Convert angle to degrees, and offset by 90 degrees for this application
    if degrees:
        angle_deg = math.degrees(angle_rad) - 90.0

        # Ensure the angle is within [0, 360] degrees
        if angle_deg < 0:
            angle_deg += 360

        return angle_deg
    else:
        return angle_rad



from artof_utils.schemas.affine import Affine
from pydantic import BaseModel, ConfigDict, model_serializer
import numpy as np
from typing import Any, Dict


class State(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    R: np.ndarray
    T: np.ndarray
    T_cov: np.ndarray
    R_cov: np.ndarray

    @model_serializer
    def ser_model(self) -> Dict[str, Any]:
        return {'R': self.R.tolist(), 'T': self.T.tolist(), 'T_cov': self.T_cov.tolist(), 'R_cov': self.R_cov.tolist()}

    def __init__(self, j=None):
        if j is None:
            j = {'R': [0, 0, 0], 'T': [0, 0, 0], 'T_cov': [0, 0, 0], 'R_cov': [0, 0, 0]}

        super().__init__(R=np.array(j['R']),
                         T=np.array(j['T']),
                         T_cov=np.array(j['T_cov']),
                         R_cov=np.array(j['R_cov']))

    def __eq__(self, other):
        return (np.array_equal(self.R, other.R) and
                np.array_equal(self.T, other.T) and
                np.array_equal(self.T_cov, other.T_cov) and
                np.array_equal(self.R_cov, other.R_cov))

    def __str__(self):
        return f"R: {self.R.tolist()}\r\n" \
               f"T: {self.T.tolist()}\r\n" \
               f"R_cov: {self.R_cov.tolist()}\r\n" \
               f"T_cov: {self.T_cov.tolist()}\r\n"

    def as_affine(self):
        """
        Creates an affine from this state
        :return: affine
        """
        return Affine().from_vectors(self.T, self.R)

    def to_json(self):
        return {
            'R': self.R.tolist(),
            'T': self.T.tolist(),
            'T_cov': self.T_cov.tolist(),
            'R_cov': self.R_cov.tolist()
        }



from artof_utils.helpers import traject
from tests.visual_support import plot_side_by_side
import pandas as pd


file_name = "files/traject/complex.csv"
distance = 3  # -2, 2, 3
side = 'end'  # 'begin', 'end'

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    df = pd.read_csv(file_name)
    df = df.apply(pd.to_numeric, errors='coerce')

    # Calculate orientation of points
    path = [[xy[0], xy[1]] for xy in zip(df['X'], df['Y'])]
    path_shifted = traject.perform(traject.Operation.ADD, path, distance=distance, side=side)

    plot_side_by_side({"original": path, "add %.2f" % distance: path_shifted})



from artof_utils.helpers import polygon
from tests.visual_support import plot_on_top


distance = 0.1  # -2, 2, 3

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    multipolygon_shape = [[[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]], [[2, 2], [3, 2], [3, 3], [2, 3]]]

    # Calculate orientation of points
    multipolygon_buffer = polygon.buffer(multipolygon_shape, distance)

    plot_data = dict()
    for i, polygon_original in enumerate(multipolygon_shape):
        plot_data["original %d" % i] = polygon_original
    for i, polygon_buffer in enumerate(multipolygon_buffer):
        plot_data["buffer %d" % i] = polygon_buffer

    plot_on_top(plot_data)



import json
from os import listdir
import re
import numpy as np
from shutil import copytree, rmtree, move
from os import path, makedirs
from pydantic import BaseModel, ConfigDict
from typing import Any, Union
import geopandas as gpd

from artof_utils.schemas.task import Task
from artof_utils.schemas.task import TaskInfo, HitchType, HitchName
import artof_utils.paths as paths
from artof_utils.redis_instance import redis_server
from artof_utils.shapefile import Shapefile, GeomType
from artof_utils.helpers import shape as shp


def get_field_names() -> list:
    assert path.exists(paths.fields), 'Field path %s does not exist.' % paths.fields
    field_names_ = listdir(paths.fields)
    assert len(field_names_) > 0, 'The folder %s does not contain any fields' % paths.fields

    return field_names_


def get_current_field_name():
    # Assert the active_field exists
    field_name = redis_server.get_value('pc.field.name')

    if not field_name:
        field_name = get_field_names()[0]
        redis_server.set_value('pc.field.name', field_name)

    return field_name


class Field(BaseModel):
    pass


class FieldInfo(BaseModel):
    name: str
    tasks: list[TaskInfo]

    @property
    def context(self):
        return self.model_dump(exclude_none=True, exclude_defaults=True)


class Fields(BaseModel):
    current_field: str = ""
    fields: list[str] = []

    def __init__(self):
        """
        The calculations are performed on the dictionary objects whereas it is not necessary to interpret the Field
        :return: json object listing the fields and there distance to current_state in [m]
        """
        super().__init__(current_field=get_current_field_name(), fields=get_field_names())

    def select_field(self, field_name):
        assert field_name in self.fields, f"Field {field_name} does not exist."

        if field_name != self.current_field:
            redis_server.set_n_values({'pc.field.name': field_name, 'pc.field.updated': True})

    def delete_field(self, field_name):
        assert field_name != self.current_field, f"Field {field_name} is currently active."

        if self.exists(field_name):
            field_path = path.join(paths.base, "field", field_name)
            rmtree(field_path, ignore_errors=True)

    def duplicate_field(self, field_name) -> Field:
        assert field_name in self.fields, f"Field {field_name} does not exist."

        field_path = path.join(paths.base, "field", field_name)
        new_field_name = f"{field_name}_copy"
        new_field_path = path.join(paths.base, "field", new_field_name)
        copytree(str(field_path), str(new_field_path), dirs_exist_ok=True)

        return Field(new_field_name)

    def exists(self, field_name):
        return field_name in self.fields

    @property
    def context(self):
        return self.model_dump()


class Field(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    field_path: str
    geofence_path: str
    traject_path: str
    tasks_path: str
    info_file_path: str

    info: FieldInfo

    shp_traject: Shapefile
    shp_geofence: Shapefile

    tasks: list[Task] = []

    def __init__(self, name):
        field_path_ = path.join(str(paths.fields), name)
        geofence_path_ = path.join(str(field_path_), "geofence")
        traject_path_ = path.join(str(field_path_), "traject")
        tasks_path_ = path.join(str(field_path_), "tasks")
        info_file_path_ = path.join(str(field_path_), "info.json")

        # Process info
        update_info = False
        if path.exists(info_file_path_):
            with open(info_file_path_, 'r') as json_file:
                j = json.load(json_file)
                field_info_: FieldInfo = FieldInfo.model_validate(j)
                # overwrite name from json if name is not equal
                if field_info_.name != name:
                    field_info_.name = name
                    update_info = True
        else:
            field_info_ = FieldInfo(name=name, tasks=[])
            update_info = True

        # Process tasks
        makedirs(tasks_path_, exist_ok=True)
        tasks_: list[Task] = []
        for task_info in field_info_.tasks:
            task_path = path.join(tasks_path_, task_info.name)
            tasks_.append(Task(task_path, task_info))

        if update_info:
            with open(info_file_path_, 'w') as json_file:
                json.dump(field_info_.context, json_file, indent=4)

        # Call super class
        super().__init__(name=name,
                         field_path=field_path_,
                         geofence_path=geofence_path_,
                         traject_path=traject_path_,
                         tasks_path=tasks_path_,
                         info_file_path=info_file_path_,
                         info=field_info_,
                         shp_traject=Shapefile(traject_path_),
                         shp_geofence=Shapefile(geofence_path_),
                         tasks=tasks_
                         )

    def rename(self, new_name):
        # check if new_name exists
        assert not Fields().exists(new_name), f"Field {new_name} already exists."

        # update info
        self.info.name = new_name
        with open(self.info_file_path, 'w') as json_file:
            json.dump(self.info.context, json_file, indent=4)
        # move files
        new_path = path.join(paths.fields, new_name)
        move(self.field_path, str(new_path))

        return Field(new_name)

    @property
    def context(self):
        traject_geometry_ = self.shp_traject.context
        geofence_geometry_ = self.shp_geofence.context
        task_geometries_ = [task.context for task in self.tasks]

        # j_traject = redis_server.get_json_value('traject')
        # if j_traject:
        #     wgs84_crs = 'EPSG:4326'  # WGS 84
        #     input_crs = 'EPSG:%d' % self.shp_traject.gdf.crs.to_epsg()
        #
        #     traject_skeleton_xy = [[point['x'], point['y']] for point in j_traject['skeleton']]
        #     traject_skeleton_latlng = shp.transform_crs(input_crs, wgs84_crs, traject_skeleton_xy)
        #
        #     traject_corners_xy = [[point['x'], point['y']] for point in j_traject['corners']]
        #     traject_corners_latlng = shp.transform_crs(input_crs, wgs84_crs, traject_corners_xy)
        #
        #     traject_geometry_['skeleton'] = {'xy': traject_skeleton_xy, 'latlng': traject_skeleton_latlng}
        #     traject_geometry_['corners'] = {'xy': traject_corners_xy, 'latlng': traject_corners_latlng}

        task_dict = dict()
        for task in sorted(self.tasks):
            task_dict[task.name] = task.context

        field_data = {
            'traject_geometry': traject_geometry_,
            'geofence_geometry': geofence_geometry_,
            'task_geometries': task_geometries_,
            'field_json': json.dumps({'name': self.name, 'traject': traject_geometry_,
                                      'geofence': geofence_geometry_, 'tasks': task_dict})
        }

        return field_data

    def update_info(self):
        self.info.tasks = [task.task_info for task in self.tasks]

        with open(self.info_file_path, 'w') as json_file:
            json.dump(self.info.context, json_file, indent=4)

    def update_geofence(self, geometries: Union[list, np.array, gpd.GeoDataFrame] | None = None, epsg: int = 0):
        self.shp_geofence.update(geometries, GeomType.POLYGON, epsg=epsg)

    def update_traject(self, geometries: Union[list, np.array, gpd.GeoDataFrame] | None = None, epsg: int = 0):
        self.shp_traject.update(geometries, GeomType.LINESTRING, epsg=epsg)

    def update_task(self, task_name, geometries: Union[list, np.array, gpd.GeoDataFrame] | None = None,
                    task_info: TaskInfo | None = None, epsg: int = 0):
        task = self.get_task(task_name)
        assert task is not None, f"Task {task_name} does not exist."

        if task_info is not None:
            task.update_info(task_info)
            self.update_info()

        if geometries is not None:
            task.update(geometries, epsg=epsg)

    def add_new_task(self):
        # Extract basename
        regex_pattern = r'\d+'
        base_names = set()
        task_numbers = set()
        for task in self.tasks:
            name = task.name
            result_name = re.sub(regex_pattern, '', name)
            result_number = re.search(regex_pattern, name)
            base_names.add(result_name)
            task_numbers.add(int(result_number.group()))

        assert len(base_names) <= 1, "All tasks must have the same basename"
        if len(base_names) == 0:
            base_name = 'Task'
        else:
            base_name = base_names.pop()
        task_numbers = set(list(range(1, 100))) - set(task_numbers)

        # Create new task
        new_task_name = base_name + str(min(task_numbers))

        # Add new task
        new_task_info = TaskInfo(name=new_task_name, type=HitchType.HITCH, hitch=HitchName.HITCH_FB)
        self.add_task(new_task_info)

    def add_task(self, task_info: TaskInfo, geometries: Union[list, np.array, gpd.GeoDataFrame] | None = None):
        task = Task(task_path=path.join(self.tasks_path, task_info.name), task_info=task_info)
        if geometries is not None:
            task.update(geometries)
        self.tasks.append(task)  # Add task to list
        self.update_info()

    def remove_task(self, task_name):
        task = self.get_task(task_name)
        assert task is not None, f"Task {task_name} does not exist."
        task.delete()  # Remove shapefiles from storage
        self.tasks.remove(task)  # Remove task from list
        self.update_info()

    def get_task(self, task_name):
        for task in self.tasks:
            if task.name == task_name:
                return task
        return None



from artof_utils.helpers import traject
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from shapely.geometry import Point

file_name = "files/traject/complex.csv"

if __name__ == '__main__':
    # Example usage:
    # Read the CSV file into a DataFrame
    df = pd.read_csv(file_name)
    df = df.apply(pd.to_numeric, errors='coerce')
    points = [Point(xy) for xy in zip(df['X'], df['Y'])]

    # Calculate orientation of points
    rows = traject.__extract_rows(points)

    # Plot result
    lw = 2
    colors = ['red', 'orange', 'green', 'blue']
    i = 0
    # Draw orientation vectors
    for row in rows:
        x_values = [traject_point.coordinate.x for traject_point in row]
        y_values = [traject_point.coordinate.y for traject_point in row]
        row_color = colors[i % len(colors)]
        plt.scatter(x_values, y_values, color=row_color, marker=".", label="row %d" % i)

        for traject_point in row:
            dx = np.cos(traject_point.orientation)
            dy = np.sin(traject_point.orientation)
            plt.arrow(traject_point.coordinate.x, traject_point.coordinate.y, dx, dy, head_width=0.2,
                      head_length=0.2, fc=row_color, ec=row_color)
        i += 1

    plt.legend(loc="lower left")
    plt.xlabel("Input")
    plt.ylabel("Response")
    plt.show()



import numpy as np
from scipy.spatial.transform import Rotation


class Affine:

    def __init__(self):
        self.a = np.identity(4)

    def get(self):
        return self.a

    def get_rotation_matrix(self):
        return self.a[0:3, 0:3]

    def get_translation_vector(self):
        return self.a[0:3, 3].transpose()

    def from_vectors(self, t, r):
        """
        Creates affine from two vectors
        :param t: vector (x, y, z) [m]
        :param r: (pitch, roll, yaw) [°]
        :return: affine
        """

        self.a[0:3, 3] = t.transpose()
        self.a[0:3, 0:3] = Rotation.from_euler('xyz', r, degrees=True).as_matrix()
        return self

    def to_vectors(self):
        """
        creates affine in vector format
        :return: t_(x, y, z) [m], r_(pitch, roll, yaw) [°]
        """
        t_ = self.get_translation_vector()
        r_m = self.get_rotation_matrix()
        r_ = Rotation.from_matrix(r_m).as_euler('xyz', degrees=True)
        return t_, r_



from pydantic import BaseModel, ConfigDict
from artof_utils.schemas.settings import HitchType, HitchName
from artof_utils.shapefile import Shapefile, GeomType
from artof_utils.schemas.implement import Implement
from typing import Optional, Union
from os import path, makedirs, removedirs
import numpy as np
import geopandas as gpd
from shutil import rmtree


class TaskInfo(BaseModel):
    name: str
    type: HitchType
    hitch: HitchName
    implement: str = ''


class Task(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    task_path: str
    type: HitchType
    hitch: HitchName

    implement: Optional[Implement] = None
    shp_task: Shapefile

    def __init__(self, task_path: str, task_info: TaskInfo):
        """
        Initializes the Task object.

        Args:
            task_path (str): The path to the task.
            task_info (TaskInfo): Information about the task.

        Returns:
            None
        """
        # Create the path if it does not exist
        if not path.exists(task_path):
            makedirs(task_path, exist_ok=True)

        name_ = task_info.name
        shp_task_ = Shapefile(task_path)
        implement_ = Implement.load(task_info.implement) if task_info.implement else None

        super().__init__(name=name_,
                         task_path=task_path,
                         type=task_info.type,
                         hitch=task_info.hitch,
                         implement=implement_,
                         shp_task=shp_task_)

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name

    def delete(self):
        rmtree(self.task_path)

    def save(self):
        makedirs(self.task_path, exist_ok=True)
        self.shp_task.save()

    def update(self, geometries: Union[list, np.array, gpd.GeoDataFrame], epsg: int = 0):
        if isinstance(geometries, gpd.GeoDataFrame):
            self.shp_task.update(geometries)
        else:
            if self.type in [HitchType.HITCH, HitchType.CONTINUOUS, HitchType.CARDAN]:
                geom_type = GeomType.POLYGON
            elif self.type in [HitchType.DISCRETE, HitchType.INTERMITTENT]:
                geom_type = GeomType.POINT
            else:
                geom_type = GeomType.POLYGON

            self.shp_task.update(geometries, geom_type, epsg=epsg)

    def update_info(self, task_info: TaskInfo):
        self.type = task_info.type
        self.hitch = task_info.hitch
        self.implement = Implement.load(task_info.implement) if task_info.implement else None

    @property
    def task_info(self):
        return TaskInfo(name=self.name,
                        type=self.type,
                        hitch=self.hitch,
                        implement=self.implement.name if self.implement else '')

    @property
    def context(self):
        return {
            'name': self.name,
            'type': self.type.value,
            'hitch': self.hitch.value,
            'implement': self.implement.name if self.implement else '',
            'geometry': self.shp_task.context
        }



from unittest import TestCase
from artof_utils.schemas.field import Fields, Field
from artof_utils.schemas.task import TaskInfo, HitchType, HitchName
from artof_utils.shapefile import GeomType
import json
from os import path
import numpy as np


class TestFields(TestCase):
    def test_load(self):
        # Arrange
        # Load existing field
        example = Field('example')
        # Act
        context = example.context
        # Assert
        self.assertFalse(context['traject_geometry']['hasZ'])

    def test_load_1(self):
        # Arrange
        # Load existing field
        example = Field('example1')
        # Act
        context = example.context
        # Assert
        self.assertFalse(context['geofence_geometry']['hasZ'])

    def test_load_2(self):
        # Arrange
        # Load existing field
        example = Field('test_tv115_drive_in')
        # Act
        context = example.context
        # Assert
        self.assertFalse(context['geofence_geometry']['hasZ'])

    def test_load_3(self):
        # Arrange
        # Load existing field
        example = Field('test_veld_tv115')
        # Act
        context = example.context
        # Assert
        self.assertFalse(context['geofence_geometry']['hasZ'])


    def test_load_discrete(self):
        # Arrange
        # Load existing field
        example = Field('example_discrete')
        task_info = TaskInfo(name='Task1', type='discrete', implement='penetrometer', hitch='RB')
        geometries = [[50.98025314667279, 3.774201817882501], [50.98040955139628, 3.7740730721325884], [50.98068562130093, 3.7738446381681716], [50.98076002267542, 3.773785085870838], [50.98073868953959, 3.7738332178084626], [50.980462629247924, 3.774060204906186], [50.98024215091789, 3.7742410751408064], [50.98028717919914, 3.774227893352885], [50.98050291227711, 3.7740737199976837], [50.98049982633975, 3.7741257733643296], [50.980606381292304, 3.7741087297280322], [50.98043429187918, 3.7742259980617856], [50.980598264702074, 3.7740926738451805], [50.980496736769894, 3.774247299568263]]
        example.update_task('Task1', geometries, task_info, epsg=4326)
        # Act
        context = example.context
        task_context = example.tasks[0].context
        # Assert
        self.assertTrue('points' in context['task_geometries'][0]['geometry'])
        self.assertTrue(example.tasks[0].shp_task.geom_type == GeomType.POINT)
        self.assertTrue(task_context['implement'] == 'penetrometer')

    def test_select_field(self):
        # Arrange
        # Create new field
        new_field = Field('field_new')
        # Act
        Fields().select_field(new_field.name)
        # Assert
        self.assertEqual(Fields().current_field, new_field.name)
        self.assertTrue(path.exists(new_field.info_file_path))
        from artof_utils.redis_instance import redis_server
        # Set 'field_updated' back to False. This is to indicate that the field was updated.
        # If no test, the cpp 'FieldManager' would otherwise do this
        redis_server.set_value("pc.field.updated", False)
        # select example again
        Fields().select_field('example')

    def test_delete_field(self):
        # Arrange
        # select field example
        Fields().select_field('example')
        # Create new field
        new_field = Field('field_new')
        # Act
        Fields().delete_field(new_field.name)
        # Assert
        self.assertFalse(path.exists(new_field.info_file_path))

    def test_duplicate_field(self):
        # Arrange
        fields = Fields()
        Fields().select_field('example_discrete')
        # Act
        field_new = fields.duplicate_field(fields.current_field)
        fields_new = Fields()
        # Assert
        self.assertIn('_copy', field_new.name)
        self.assertIn(field_new.name, fields_new.fields)
        with open(field_new.info_file_path, 'r') as f:
            info = json.load(f)
            self.assertEqual(field_new.name, info['name'])

    def test_create_new_field(self):
        # Arrange
        geofence_coords = [[[0.0, 0.0], [0.0, 100.00], [100.0, 100.0], [100.0, 0.0]]]
        traject_coords = np.array([[50.0, 5.0], [50.0, 95.00], [55.0, 95.0], [55.0, 5.0]])
        task1_coords = np.array([[10.0, 10.0], [10.0, 90.00], [90.0, 90.0], [90.0, 10.0], [10.0, 10.0]])
        task2_coords = np.array([[[10.0, 10.0], [10.0, 90.00], [90.0, 90.0], [90.0, 10.0]]])

        field_new = Field('example_new')
        # Be sure no tasks in the system for this field
        task_names = [task.name for task in field_new.tasks]
        for task_name in task_names:
            field_new.remove_task(task_name)

        # Act
        field_new.update_traject(traject_coords)
        field_new.update_geofence(geofence_coords)
        field_new.add_task(TaskInfo(name='task1', type=HitchType.HITCH, hitch=HitchName.HITCH_FB), task1_coords)
        field_new.add_task(TaskInfo(name='task2', type=HitchType.CONTINUOUS, hitch=HitchName.HITCH_FB), task2_coords)
        field_new.add_new_task()
        # Assert
        self.assertTrue(path.exists(field_new.shp_geofence.file_path))
        self.assertTrue(path.exists(field_new.shp_traject.file_path))
        for task_name in ['task1', 'task2']:
            task = field_new.get_task(task_name)
            self.assertTrue(path.exists(task.shp_task.file_path))
            self.assertEqual(task.shp_task.get_geom_type(), GeomType.POLYGON)

        task_name = [task.name for task in field_new.tasks]
        self.assertEqual(set(task_name), {'task1', 'task2', 'task3'})

    def test_rename_field(self):
        # Arrange
        orig_name = 'example_name_new'
        new_name = 'example_name_renamed_new'
        Fields().delete_field(new_name)
        # Act
        field = Field(orig_name)
        renamed_field = field.rename(new_name)
        # Assert
        self.assertEqual(renamed_field.name, new_name)



from enum import Enum
import numpy as np
from shapely.geometry import Polygon
from artof_utils.helpers import array


class Operation(Enum):
    BUFFER = 'buffer'

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value


def create(transform: np.array, width: float = 0, height: float = 0, up: float = 0, down: float = 0):
    assert width != 0, 'Width must be positive'
    assert height != 0 or up != 0 or down != 0, 'Height must set or up and down must be set'

    if height != 0:
        up = height / 2
        down = - height / 2

    # Create polygon
    point_ll = [- (width / 2), down, 0, 1.0]
    point_lr = [+ (width / 2), down, 0, 1.0]
    point_ur = [+ (width / 2), up, 0, 1.0]
    point_ul = [- (width / 2), up, 0, 1.0]
    points = np.array([point_ll, point_lr, point_ur, point_ul, point_ll])

    co_xyzw = np.dot(points, transform.T)
    return [[p[0], p[1]] for p in co_xyzw]


# Operations
def perform(operation: Operation, data, **kwargs):
    if operation == Operation.BUFFER:
        return buffer(data, **kwargs)
    return data


def buffer(data: list, distance: float) -> list:
    depth = array.get_depth(data)

    if depth > 2:
        return [buffer(polygon, distance) for polygon in data]
    else:
        # Make sure we have a closed polygon
        if np.all(np.isclose(data[0], data[-1], atol=1e-4)):
            data[-1] = data[0]
        else:
            data.append(data[0])

        polygon = Polygon(data)
        polygon_buffered = polygon.buffer(distance, cap_style=3, join_style=2)

        return list(polygon_buffered.exterior.coords)



import numpy as np


def __numeric(value):
    """
    Check if value is numeric
    :param value: Value to check
    :return: true if value is numeric, false otherwise
    """
    return (isinstance(value, (int, float)) or
            isinstance(value, (np.float32, np.float64, np.float16)) or
            isinstance(value, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64)))


def __get_depth_list(data):
    """
    Get depth of list
    :param data: List
    :return: depth of list
    """
    if not isinstance(data, list):
        return 0
    elif len(data) == 0:
        return 1
    else:
        # Item is a list, numpy array, scalar, None or object (e.g. TrajectPoint)
        return 1 + max(0 if (__numeric(item) or item is None) else
                       __get_depth_list(item) if isinstance(item, list) else
                       __get_depth_np(item) if isinstance(item, np.ndarray) else
                       1
                       for item in data)


def __get_depth_np(data):
    """"
    Get depth of numpy array.
    Numpy arrays require that every dimension is uniform,
    meaning all the nested sequences must have the same length to form a multidimensional array.
    :param data: numpy array
    :return: depth of array
    """
    depth = 0
    while isinstance(data, np.ndarray):
        depth += 1
        data = data[0] if data.size > 0 else None
    return depth


def get_depth(arr):
    """
    Get depth of array (public function)
    :param arr: array
    :return: depth of array
    """
    depth_lst = __get_depth_list(arr)
    depth_np = __get_depth_np(arr)
    return max(depth_lst, depth_np)
    #if isinstance(arr, np.ndarray) : return __get_depth_np(arr)
    #elif isinstance(arr, list) : return __get_depth_list(arr)
    #else : return 0




import json
from typing import Optional
from pydantic import BaseModel
from artof_utils.schemas.settings import Transform, HitchType
import artof_utils.paths as paths
from os import path


class Section(BaseModel):
    id: str = ""
    width: float = 0.0
    up: float = 0.03
    down: float = 0.03
    link_length: Optional[float] = None
    repeats: Optional[int] = None
    offset: Optional[float] = None
    transform: Transform


class Implement(BaseModel):
    name: str
    on_taskmap: Optional[bool] = True
    types: list[HitchType] = []
    sections: list[Section] = []

    @staticmethod
    def load(name: str):
        implement_file_path = path.join(paths.implements, '%s.json' % name)
        if path.exists(implement_file_path):
            with open(implement_file_path, 'r') as json_file:
                j = json.load(json_file)
                return Implement(**j)
        else:
            return Implement(name=name)

    @property
    def context(self):
        data = self.model_dump(exclude_none=True)
        data['types'] = [t.value for t in self.types] if self.types else []
        return data



from artof_utils.helpers import traject
from tests.visual_support import plot_side_by_side
import pandas as pd


file_name = "files/traject/complex.csv"
row_number = [2, 3]  # 1, 2, [0, 1], [0, 2]

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    df = pd.read_csv(file_name)
    df = df.apply(pd.to_numeric, errors='coerce')

    # Calculate orientation of points
    path = [[xy[0], xy[1]] for xy in zip(df['X'], df['Y'])]
    path_shifted = traject.perform(traject.Operation.REMOVE, path, row_number=row_number)

    plot_side_by_side({"original": path, "remove": path_shifted})



from pydantic import BaseModel
from typing import Optional
from artof_utils.schemas.settings import Hitch
from artof_utils.redis_instance import redis_server
from copy import deepcopy


class Hitches(BaseModel):
    hitches: Optional[list[Hitch]] = []

    redis_variables: list[str]

    def __init__(self):
        super().__init__(redis_variables=[])

    def add_hitches(self, hitches: list[Hitch]):
        self.hitches = deepcopy(hitches)

        # Add hitch redis variables
        for hitch in self.hitches:
            hitch_name = 'plc.control.hitch_' + hitch.name.lower()
            self.redis_variables.append(hitch_name + '.activate')
            self.redis_variables.append(hitch_name + '.active_discrete')
            self.redis_variables.append(hitch_name + '.active_continuous')
            self.redis_variables.append(hitch_name + '.active_cardan')
            self.redis_variables.append(hitch_name + '.setpoint')
            for i in range(32):
                self.redis_variables.append(hitch_name + '.activate_sections.%d' % i)

        self.update()

    def update(self):
        d = redis_server.get_n_values(self.redis_variables)

        # Update hitches
        for hitch in self.hitches:
            hitch_name = 'plc.control.hitch_' + hitch.name.lower()
            hitch.activate = d[hitch_name + '.activate']
            hitch.start_discr_impl = d[hitch_name + '.active_discrete']
            hitch.activate_cont_impl = d[hitch_name + '.active_continuous']
            hitch.activate_cardan = d[hitch_name + '.active_cardan']
            hitch.setpoint = d[hitch_name + '.setpoint']
            hitch.float = hitch.setpoint == 99
            for i in range(32):
                variable_name = hitch_name + '.activate_sections.%d' % i
                hitch.activate_section[i] = d[variable_name]

    @staticmethod
    def change(hitch_setpoints: dict):
        d = {}
        for name, setpoint in hitch_setpoints.items():
            hitch_name = 'plc.control.hitch_' + name.lower()
            d[hitch_name + '.setpoint'] = setpoint
        redis_server.set_n_values(d)

    @property
    def context(self):
        return self.model_dump(exclude={'redis_variables'})





# Gps
def parse_gps_fix_number(gps_fix_number):
    """
    Parse GPS fix number
    :param gps_fix_number: GPS fix number
    :return: GPS fix number description
    """
    if gps_fix_number == 0:
        return 'Not valid'
    elif gps_fix_number == 1:
        return 'GPS Fix'
    elif gps_fix_number == 2:
        return 'Differential GPS'
    elif gps_fix_number == 3:
        return 'Not applicable'
    elif gps_fix_number == 4:
        return 'RTK Fix'
    elif gps_fix_number == 5:
        return 'RTK Float'
    else:
        return 'Not valid'



import geopandas as gpd
from os import path, makedirs
from glob import glob
import numpy as np
from typing import Union
from shapely.geometry import Point, Polygon, LineString
from enum import Enum
from pyproj import CRS
from artof_utils.helpers import array
from artof_utils.helpers import shape as shp
from artof_utils.schemas.settings import load_settings


class GeomType(str, Enum):
    POINT = 'Point'
    MULTIPOINT = 'MultiPoint'
    LINESTRING = 'LineString'
    POLYGON = 'Polygon'


class Shapefile:
    def __init__(self, folder_path):
        self.empty = True
        self.geom_type = None
        shape_files = glob(path.join(folder_path, '*.shp'))
        if len(shape_files) == 0:
            # Create a new empty shapefile as it does not exist
            self.file_path = path.join(folder_path, '%s.shp' % path.basename(folder_path))
            settings = load_settings()
            self.gdf = gpd.GeoDataFrame(geometry=[Point(0, 0)], crs=CRS('EPSG:326%d' % settings.gps.utm_zone))
            makedirs(folder_path, exist_ok=True)
            self.save()
        else:
            # Read the shapefile as it exists
            self.file_path = shape_files[0]
            self.gdf = gpd.read_file(self.file_path)
            self.geom_type = self.gdf.geom_type.iloc[0] if len(self.gdf.geom_type) else None
            self.empty = (len(self.gdf.geometry) == 1 and
                          self.geom_type == GeomType.POINT and
                          list(self.gdf.geometry[0].coords) == [(0.0, 0.0)]) or self.geom_type is None

    @property
    def context(self):
        r = dict()

        r['empty'] = self.empty
        r['hasZ'] = bool(self.gdf.has_z.iloc[0]) if len(self.gdf.has_z) else False
        r['wkid'] = self.gdf.crs.to_epsg()

        wgs84_crs = 'EPSG:4326'  # WGS 84
        input_crs = 'EPSG:%d' % r['wkid']

        if self.geom_type == GeomType.POINT:
            if r['hasZ']:
                r['points'] = self.gdf.geometry.apply(lambda point: [point.x, point.y, point.z]).to_list()
            else:
                r['points'] = self.gdf.geometry.apply(lambda point: [point.x, point.y]).to_list()
            r['latlng'] = shp.transform_crs(input_crs, wgs84_crs, r['points'])
        elif self.geom_type == GeomType.MULTIPOINT:
            if r['hasZ']:
                np_points = np.array(self.gdf.geometry.apply(
                    lambda multipoint: [[point.x, point.y, point.z] for point in multipoint.geoms]).to_list())
                np_points = np_points.squeeze(axis=1)
            else:
                np_points = np.array(self.gdf.geometry.apply(
                    lambda multipoint: [[point.x, point.y] for point in multipoint.geoms]).to_list())
                np_points = np_points.squeeze(axis=1)
            r['points'] = [list(point) for point in np_points]
            r['latlng'] = shp.transform_crs(input_crs, wgs84_crs, r['points'])
        elif self.geom_type == GeomType.LINESTRING:
            if r['hasZ']:
                r['paths'] = self.gdf.geometry.apply(
                    lambda linestring: [[x, y, z] for x, y, z in linestring.coords]).to_list()
            else:
                r['paths'] = self.gdf.geometry.apply(
                    lambda linestring: [[x, y] for x, y in linestring.coords]).to_list()
            r['latlng'] = [shp.transform_crs(input_crs, wgs84_crs, path_) for path_ in r['paths']]
        elif self.geom_type == GeomType.POLYGON:
            if r['hasZ']:
                r['rings'] = self.gdf.geometry.apply(
                    lambda polygon: [[x, y, z] for x, y, z in polygon.exterior.coords]).to_list()
            else:
                r['rings'] = self.gdf.geometry.apply(
                    lambda polygon: [[x, y] for x, y in polygon.exterior.coords]).to_list()
            r['latlng'] = [shp.transform_crs(input_crs, wgs84_crs, ring_) for ring_ in r['rings']]
        else:
            r['latlng'] = []

        return r

    def update(self, geometries: Union[list, np.array, gpd.GeoDataFrame], geom_type: GeomType = None, epsg: int = 0):
        if isinstance(geometries, gpd.GeoDataFrame):
            # Process gdf as geometry
            assert geometries.crs, 'No crs in shapefile'
            if geom_type:
                assert geometries.geom_type[0] == geom_type, 'Only %s geometries are supported' % geom_type

            self.gdf = geometries
            self.save()
        else:
            # Process list and unions as geometry
            assert self.gdf.crs, 'No crs in shapefile'

            if epsg:
                # Convert to this crc
                input_crs = 'EPSG:%d' % epsg
                gdf_crs = 'EPSG:%d' % self.gdf.crs.to_epsg()
                if input_crs != gdf_crs:
                    geometries = shp.transform_crs(input_crs, gdf_crs, geometries)

            depth = array.get_depth(geometries)

            if not geom_type:
                # Use geometry type of first geometry
                assert len(self.gdf) > 0 and self.geom_type is not None, 'No data in shapefile'
                geom_type = self.geom_type
            else:
                self.geom_type = geom_type

            new_geom = None
            if geom_type == GeomType.POINT or geom_type == GeomType.MULTIPOINT:
                new_geom = [Point(geom[0], geom[1]) for geom in np.array(geometries).squeeze()]
            elif geom_type == GeomType.LINESTRING:
                if depth == 2:
                    geometries = np.expand_dims(geometries, axis=0) if isinstance(geometries, np.ndarray) else [
                        geometries]
                # Create array of rings when only one ring is provided
                new_geom = [LineString(geom) for geom in geometries]
            elif geom_type == GeomType.POLYGON:
                # Create array of rings when only one ring is provided
                if depth == 2:
                    geometries = (np.expand_dims(geometries, axis=0) if isinstance(geometries, np.ndarray) else
                                  [[np.array(geom) for geom in geometries]])
                elif depth == 3:
                    if isinstance(geometries, list):
                        geometries = [np.array(geom) for geom in geometries]
                elif depth == 4:
                    geometries = geometries[0]
                else:
                    raise ValueError('unsupported depth of array')

                # Check if rings are closed and if not close them
                for geom in geometries:
                    assert len(geom) >= 3, 'Empty polygon'
                    if np.allclose(geom[0], geom[-1]):
                        geom = np.vstack((geom, geom[0]))  # update the geometry
                new_geom = [Polygon(geom) for geom in geometries]

            if new_geom:
                self.gdf = gpd.GeoDataFrame(geometry=new_geom, crs=self.gdf.crs)
                self.save()

        self.empty = False

    def save(self, other_folder_path=None):
        if other_folder_path:
            makedirs(other_folder_path, exist_ok=True)
            save_file_path = path.join(other_folder_path, path.basename(self.file_path))
        else:
            save_file_path = self.file_path

        self.gdf.to_file(save_file_path)

    def get_geom_type(self):
        if len(self.gdf) == 0:
            return None

        return self.gdf.geom_type[0]



from artof_utils.helpers import polygon
from tests.visual_support import plot_on_top
import numpy as np


distance = -0.1  # -2, 2, 3

if __name__ == '__main__':
    # Read the CSV file into a DataFrame
    polygon_shape = [[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]]

    # Calculate orientation of points
    commands = {'distance': distance}
    polygon_buffer = polygon.perform(polygon.Operation('buffer'), polygon_shape, **commands)

    plot_on_top({"original": polygon_shape, "buffer": polygon_buffer})



from unittest import TestCase
import numpy as np
import pandas as pd
from shapely.geometry import Point
from artof_utils.helpers import traject


class TestHelpersTraject(TestCase):
    def test_reverse_point_array_list_1(self):
        points = [[0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]]
        reversed_points = traject.reverse(points)
        self.assertEqual(reversed_points, [[9, 9], [8, 8], [7, 7], [6, 6], [5, 5], [4, 4], [3, 3], [2, 2], [1, 1], [0, 0]])

    def test_reverse_points_np_1(self):
        points = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]])
        reversed_points = traject.reverse(points)
        self.assertTrue(np.array_equal(reversed_points, [[9, 9], [8, 8], [7, 7], [6, 6], [5, 5], [4, 4], [3, 3], [2, 2], [1, 1], [0, 0]]))

    def test_reverse_points_list_2(self):
        points = [[[0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]], [[0, 0], [1, 1], [2, 2], [3, 3]]]
        reversed_points = traject.reverse(points)
        self.assertEqual(reversed_points, [[[9, 9], [8, 8], [7, 7], [6, 6], [5, 5], [4, 4], [3, 3], [2, 2], [1, 1], [0, 0]], [[3, 3], [2, 2], [1, 1], [0, 0]]])

    def test_reverse_points_np_2(self):
        points = np.array([[[0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]]])
        reversed_points = traject.reverse(points)
        self.assertTrue(np.array_equal(reversed_points, [[[9, 9], [8, 8], [7, 7], [6, 6], [5, 5], [4, 4], [3, 3], [2, 2], [1, 1], [0, 0]]]))

    def test_traject_extract(self):
        file_name = "tests/files/traject/complex.csv"
        df = pd.read_csv(file_name)
        df = df.apply(pd.to_numeric, errors='coerce')
        points = [Point(xy) for xy in zip(df['X'], df['Y'])]
        rows = traject.get_rows(points)

        self.assertEqual(len(rows), 4, "Expected 4 rows, got %d" % len(rows))


from unittest import TestCase
from artof_utils.schemas.implement import Implement


class TestImplementManager(TestCase):

    def test_load_implements(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.implement import implement_manager

        # Act
        implements = implement_manager.implements

        # Assert
        self.assertGreater(len(implements), 0)

    def test_context(self):
        # Arrange
        # - Load settings done as singleton
        from artof_utils.implement import implement_manager

        # Act
        context = implement_manager.implements[0].context

        # Assert
        for t in context['types']:
            self.assertTrue(type(t) is str)
        self.assertIsNotNone(context)

    def test_get_implement(self):
        from artof_utils.implement import implement_manager
        implement = implement_manager.get_implement('new')
        context = implement.context

        self.assertTrue(context['on_taskmap'])
        self.assertTrue(len(context['sections']) == 0)

    def test_parse_implement(self):
        data = {'name': 'new', 'on_taskmap': False, 'types': ['continuous'], 'sections': [
            {'id': 'P', 'width': 0.25, 'up': 0.1, 'down': 0.1, 'transform': {'T': [0, 0, 0], 'R': [0, 0, 0]}}]}

        implement = Implement(**data)

        self.assertFalse(implement.on_taskmap)
        self.assertTrue(len(implement.sections) > 0)
        self.assertTrue('offset' not in implement.context['sections'][0])



class Singleton(type):
    """
    A metaclass that creates a Singleton instance for a class.

    The Singleton pattern ensures that a class has only one instance and provides a global point of access to it.
    This metaclass implements the pattern by intercepting calls to the class's constructor to ensure that only one
    instance is created throughout the lifetime of the application. If the class has already been instantiated,
    the original instance is returned to all subsequent callers.

    Attributes:
        _instances (dict): A class-level dictionary that maps a class to its single instance.
    """

    _instances = {}  # Stores the singleton instance for each class.

    def __call__(cls, *args, **kwargs):
        """
        This method checks if the class instance already exists in the `_instances` dictionary.
        If it does, it returns that instance. If not, it creates a new instance, stores it in
        `_instances`, and then returns it.

        :param cls: The class for which the singleton instance is being requested.
        :param *args: Positional arguments to pass to the class constructor.
        :param *kwargs: Keyword arguments to pass to the class constructor.

        :return:The singleton instance of the class.
        """
        if cls not in cls._instances:
            # If the instance does not exist, create it and store it in the _instances dict.
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from artof_utils.schemas.settings import HitchType, HitchName
from artof_utils.implement import implement_manager


def get_hitch_choices():
    return [(hitch_name.value, hitch_name.value) for hitch_name in HitchName]


def get_type_choices():
    return [(hitch_type.value, hitch_type.value) for hitch_type in HitchType]


def get_implement_choices():
    return [('none', None)] + [(implement.name, implement.name) for implement in implement_manager.implements]
